/****************************************************************************
**  SIONLIB     http://www.fz-juelich.de/jsc/sionlib                       **
*****************************************************************************
**  Copyright (c) 2008-2018                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

/*!
 * @file sion_printts.h
 * @brief Sion Time Stamp Header
 * @author Wolfgang Frings
 */

#ifndef SION_SION_PRINTTS_H
#define SION_SION_PRINTTS_H

int lsion_print_time_stamp (int rank, char *desc);
double _lsion_get_time(void);

#endif
